#include <PalmOS.h>
#include "DB.h"
#include "dcCalls.h"
#include "util.h"
#include "VfsMgr.h"


//////////////////////////STRUCTURES///////////

typedef struct{
	
	char		name[35];		//app name
	Boolean		card;			//set if it is on the card
	UInt32		crid;			//creator ID of the apps
	
}ListItem;


//////////////////////////GLOBALS///////////////

ListItem*	gListItems;
UInt32		gNumListItems;

const char*	launcherFolderName[]={"/Palm/Launcher/","/Applications/",""};
const char*	strSorting="Sorting...";
const char* strEnumerating="Enumerating...";

/////////////////////////PROTOTYPES/////////////

UInt32 listAllApps(ListItem** items);		//list all apps on device and card, return number of apps found



/////////////////////////CODE///////////////////

void about(){
	
	char ver[32];
	MemHandle mh;
	FormPtr fp;
	
	
	mh=DmGetResource('tver',1000);
	MemMove(ver,MemHandleLock(mh),sizeof(ver));
	MemHandleUnlock(mh);
	DmReleaseResource(mh);
	
	
	fp=FrmInitForm(1001);
	FrmCopyLabel(fp,8888,ver);				//set version to be displayed
	FrmDrawForm(fp);
	
	mh=DmGetResource('tAIB',1000);			//draw icon
	WinDrawBitmap(MemHandleLock(mh),4,17);
	MemHandleUnlock(mh);
	DmReleaseResource(mh);
	
	FrmDoDialog(fp);
	
	FrmDeleteForm(fp);
}


#define TEMP_DB_NAME	"tempDb"

void showStr(char* str){
	
	DmOpenRef db;
	LocalID LID;
	UInt16 attr;
	MemHandle mh;
	char* dst;
	Boolean done=false;
	
	if(errNone==DmCreateDatabase(0,TEMP_DB_NAME,'psys','TEMP',true)){
		
		LID=DmFindDatabase(0,TEMP_DB_NAME);
		
		if(LID){
			
			db=DmOpenDatabase(0,LID,dmModeReadWrite);
			
			if(db){
				
				if(errNone==DmDatabaseInfo(0,LID,NULL,&attr,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL)){
					
					attr |= dmHdrAttrRecyclable;
					
					if(errNone==DmSetDatabaseInfo(0,LID,NULL,&attr,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL)){
						
						mh=DmNewResource(db,'tSTR',1000,StrLen(str)+1);
						
						if(mh){
							
							dst=MemHandleLock(mh);
							
							if(dst){
								
								if(errNone==DmStrCopy(dst,0,str)){
									
									FrmHelp(1000);
									
									done=true;
								}
								
								MemHandleUnlock(mh);
							}
							
							DmReleaseResource(mh);
						}
					}
				}
				
				DmCloseDatabase(db);
			}
		}
		
		LID=DmFindDatabase(0,TEMP_DB_NAME);
		
		if(LID){		//still not deleted? why?
			
			DmDeleteDatabase(0,LID);
			
		}
	}
	
	if(!done){
		
		ErrAlertCustom(0,"Failed to show string using system dialog! This is bad",NULL,NULL);
		
	}
}



void stats(){		//show DynaClock statistics
	
	char* txt;
	MemHandle txtH;
	Boolean controlBus=false,controlMem=false;
	UInt32 minCpu=0xFFFFFFFF,minBus=0xFFFFFFFF,minMem=0xFFFFFFFF;
	UInt32 maxCpu=0,maxBus=0,maxMem=0;
	UInt32 t;
	UInt16 num,i;
	
	//prepare
	
	
	txtH=MemHandleNew(256);		//this allocation cannot fail. if it does, you have bigger problems then this app crashing
	txt=MemHandleLock(txtH);
	
	//generate statistics
	
	txt[0]=0;
	
	num=dcGetNumStates();
	
	StrCat(txt,"DynaClock controls: CPU");		//cpu always
	
	if(dcGetStateBus(0)){						//non-zero bus = we control bus
		
		StrCat(txt,", BUS");
		
		controlBus=true;
		
	}
	if(dcGetStateMem(0)){						//non-zero mem = we control bus
		
		StrCat(txt,", MEM");
		
		controlMem=true;
		
	}
	StrCat(txt,"\nTotal PowerStates: ");
	StrIToACat(txt,num);
	StrCat(txt,"\nAverages since last reset:\n  CPU: ");
	StrIToACat_100(txt,dcGetAvgCpuSpeed());
	StrCat(txt," MHz \n");
	
	if(controlBus){
		
		StrCat(txt,"  BUS: ");
		StrIToACat_100(txt,dcGetAvgBusSpeed());
		StrCat(txt," MHz \n");
		
	}
	if(controlMem){
		
		StrCat(txt,"  MEM: ");
		StrIToACat_100(txt,dcGetAvgMemSpeed());
		StrCat(txt," MHz \n");
		
	}
	StrCat(txt,"  PowerState: ");
	StrIToACat_100(txt,dcGetAvgPowerState());
	StrCat(txt,"\n");
	
	for(i=0;i<num;i++){
		
		t=dcGetStateCpu(i);
		
		if(t>maxCpu)	maxCpu=t;
		if(t<minCpu)	minCpu=t;
		
		if(controlBus){
			
			t=dcGetStateBus(i);
			if(t>maxBus)	maxBus=t;
			if(t<minBus)	minBus=t;
			
		}
		if(controlMem){
			
			t=dcGetStateMem(i);
			if(t>maxMem)	maxMem=t;
			if(t<minMem)	minMem=t;
			
		}
		
	}
	
	StrCat(txt,"Speed ranges:\n  CPU: ");
	StrIToACat(txt,minCpu);
	StrCat(txt,"-");
	StrIToACat(txt,maxCpu);
	StrCat(txt," MHz \n");
	
	if(controlBus){
		
		StrCat(txt,"  BUS: ");
		StrIToACat(txt,minBus);
		StrCat(txt,"-");
		StrIToACat(txt,maxBus);
		StrCat(txt," MHz \n");
		
	}
	if(controlMem){
		
		StrCat(txt,"  MEM: ");
		StrIToACat(txt,minMem);
		StrCat(txt,"-");
		StrIToACat(txt,maxMem);
		StrCat(txt," MHz \n");
		
	}
	
	
	//show them
	
	showStr(txt);
	
	MemHandleUnlock(txtH);
	MemHandleFree(txtH);
}

void listPowerStates(){
	char* txt;
	MemHandle txtH;
	UInt16 num,i;
	
	//prepare
	
	
	txtH=MemHandleNew(512);		//this allocation cannot fail. if it does, you have bigger problems then this app crashing
	txt=MemHandleLock(txtH);
	
	//generate statistics
	
	txt[0]=0;
	
	num=dcGetNumStates();
	
	StrCat(txt,"Total PowerStates: ");
	StrIToACat(txt,num);
	
	for(i=0;i<num;i++){
		
		StrCat(txt,"\n    ");
		makeStateString(txt+StrLen(txt),i,dcGetStateCpu(i),dcGetStateBus(i),dcGetStateMem(i));
		
	}
	
	//show them
	
	
	showStr(txt);
	
	MemHandleUnlock(txtH);
	MemHandleFree(txtH);
}

void drawStateLabel(FieldPtr fi,UInt16 state){
	
	MemHandle mh;
	
	mh=FldGetTextHandle(fi);
	if(!mh){
		
		mh=MemHandleNew(64);
		
	}
	else{
		
		FldSetTextHandle(fi,NULL);
		MemHandleResize(mh,64);
		
	}
	
	makeStateString(MemHandleLock(mh),state,dcGetStateCpu(state),dcGetStateBus(state),dcGetStateMem(state));
	
	MemHandleUnlock(mh);
	
	FldSetTextHandle(fi,mh);
	
	FldDrawField(fi);
}

Boolean configFormEventHandler(EventPtr e){	
	
	FormPtr fp;
	ControlPtr minS;
	ControlPtr maxS;
	FieldPtr minF;
	FieldPtr maxF;
	UInt16 val;
	Boolean controllingMin;
	
	fp=FrmGetFormPtr(1003);
	minS=FrmGetObjectPtr(fp,FrmGetObjectIndex(fp,3000));
	maxS=FrmGetObjectPtr(fp,FrmGetObjectIndex(fp,3001));
	minF=FrmGetObjectPtr(fp,FrmGetObjectIndex(fp,4000));
	maxF=FrmGetObjectPtr(fp,FrmGetObjectIndex(fp,4001));
	
	
	if(e->eType==0x7999){	
		
		CtlGetSliderValues(minS,NULL,NULL,NULL,&val);
		drawStateLabel(minF,val);
		
		CtlGetSliderValues(maxS,NULL,NULL,NULL,&val);
		drawStateLabel(maxF,val);
		
	}
	if(e->eType==ctlRepeatEvent && e->data.ctlRepeat.controlID>=3000 && e->data.ctlRepeat.controlID<=3001){
		
		controllingMin=e->data.ctlRepeat.controlID==3000;
		
		drawStateLabel(controllingMin?minF:maxF,e->data.ctlRepeat.value);
		
		if(controllingMin){
			
			CtlGetSliderValues(maxS,NULL,NULL,NULL,&val);
			if(val<e->data.ctlRepeat.value){
				
				CtlSetSliderValues(maxS,NULL,NULL,NULL,&e->data.ctlRepeat.value);
				drawStateLabel(maxF,e->data.ctlRepeat.value);
				
			}
			
		}
		else{
			
			CtlGetSliderValues(minS,NULL,NULL,NULL,&val);
			if(val>e->data.ctlRepeat.value){
				
				CtlSetSliderValues(minS,NULL,NULL,NULL,&e->data.ctlRepeat.value);
				drawStateLabel(minF,e->data.ctlRepeat.value);
				
			}
			
		}
	}
	if(e->eType==ctlSelectEvent && e->data.ctlSelect.controlID>=3000 && e->data.ctlSelect.controlID<=3001){
		
		controllingMin=e->data.ctlSelect.controlID==3000;
		
		drawStateLabel(controllingMin?minF:maxF,e->data.ctlSelect.value);
		
		if(controllingMin){
			
			CtlGetSliderValues(maxS,NULL,NULL,NULL,&val);
			if(val<e->data.ctlSelect.value){
				
				CtlSetSliderValues(maxS,NULL,NULL,NULL,&e->data.ctlSelect.value);
				drawStateLabel(maxF,e->data.ctlSelect.value);
				
			}
			
		}
		else{
			
			CtlGetSliderValues(minS,NULL,NULL,NULL,&val);
			if(val>e->data.ctlSelect.value){
				
				CtlSetSliderValues(minS,NULL,NULL,NULL,&e->data.ctlSelect.value);
				drawStateLabel(minF,e->data.ctlSelect.value);
				
			}
			
		}
	}
	return false;
}

Boolean itemConfig(ListItem* i){	//return if anythign was saved
	
	FormPtr fp;
	DmOpenRef db;
	dbRecType rec;
	UInt16 bp;
	ControlPtr min;
	ControlPtr max;
	UInt16 maxV,curV;
	
	fp=FrmInitForm(1003);
	FrmSetEventHandler(fp,&configFormEventHandler);
	
	if(i->crid==SYSTEM_CRID){
		
		CtlHideControl(FrmGetObjectPtr(fp,FrmGetObjectIndex(fp,2001)));	//hide "sys" button for "System" config
		
	}
	
	FrmCopyLabel(fp,1000,i->name);
	
	
	db=openDb();
	if(!getRecord(db,i->crid,&rec) || rec.minState==0xFFFFFFFF){
		
		getRecord(db,SYSTEM_CRID,&rec);		//get system if set to "system" or does not exist
											//due to a call to makeSysRecord() earlier, we are guaranteed that
											//System record exists
		
	}
	
	min=FrmGetObjectPtr(fp,FrmGetObjectIndex(fp,3000));
	max=FrmGetObjectPtr(fp,FrmGetObjectIndex(fp,3001));
	
	maxV=dcGetNumStates()-1;
	
	
	curV=rec.minState;
	CtlSetSliderValues(min,NULL,&maxV,NULL,&curV);
	
	curV=rec.maxState;
	CtlSetSliderValues(max,NULL,&maxV,NULL,&curV);
	
	FrmDrawForm(fp);
	
	curV=0x7999;		//a custom event
	EvtAddEventToQueue((EventPtr)&curV);	//bad but saves stack
	
	bp=FrmDoDialog(fp);
	
	if(bp==2001){		//sys
		
		rec.minState=0xFFFFFFFF;
		
	}
	
	if(bp==2000){		//ok
		
		CtlGetSliderValues(min,NULL,NULL,NULL,&curV);
		rec.minState=curV;
		
		CtlGetSliderValues(max,NULL,NULL,NULL,&curV);
		rec.maxState=curV;
		
		
	}
	
	if(bp==2000 || bp==2001){
		
		setRecord(db,i->crid,&rec);
		
	}
	
	DmCloseDatabase(db);
	
	FrmDeleteForm(fp);
	
	return ((bp==2000) || (bp==2001));
}

Boolean mainFormEventHandler(EventPtr e){
	
	if(e->eType==ctlSelectEvent){
		
		switch(e->data.ctlSelect.controlID){
			
			case 2000:
				
				about();
				return true;
				break;
			
			
			case 2001:
				
				stats();
				return true;
				break;
			
			
			case 2002:
				
				listPowerStates();
				return true;
				break;
			
			
		}
		
	}
	else if(e->eType==lstSelectEvent){
		
		FormPtr fp;
		ListPtr L;
		UInt32 sel;
		
		
		fp=FrmGetFormPtr(1000);
		
		L=FrmGetObjectPtr(fp,FrmGetObjectIndex(fp,1000));	//get list ptr
		
		sel=LstGetSelection(L);
		
		if(sel!=noListSelection){
			
			if(itemConfig(&gListItems[sel])){
				
				LstDrawList(L);		//redraw list
				
			}
			
		}
		
	}
	
	return false;
	
}

#define EXTRA_COL_POS 120

void listDrawFunc(Int16 itemNum,RectanglePtr bounds,Char **itemsText){
	
	ListItem* items=(ListItem*)itemsText;
	FontID font;
	char range[7];
	dbRecType rec;
	DmOpenRef db;
	
	
	font=FntSetFont(stdFont);
	
	WinDrawChars(items[itemNum].name,StrLen(items[itemNum].name),bounds->topLeft.x+15+2,bounds->topLeft.y);
	
	db=openDb();
	if(getRecord(db,items[itemNum].crid,&rec) && (rec.minState!=0xFFFFFFFF)){
		
		StrIToA(range,rec.minState);
		StrCat(range,"-");
		StrIToACat(range,rec.maxState);
		
	}
	else{
		
		StrCopy(range,"SYS");
		
	}
	DmCloseDatabase(db);
	
	
	WinDrawChars(range,StrLen(range),bounds->topLeft.x+EXTRA_COL_POS+1,bounds->topLeft.y);
	
	
	if(items[itemNum].card){	//draw card icon
		
		WinDrawChar(chrCardIcon,bounds->topLeft.x+1+((15-FntCharWidth(chrCardIcon))/2),bounds->topLeft.y);
		
	}
	else{						//attempt to draw app's icon
		
		DmOpenRef db;
		MemHandle mh;
		
		db=DmOpenDatabaseByTypeCreator('appl',items[itemNum].crid,dmModeReadOnly);
		
		if(db){
			
			mh=DmGet1Resource('tAIB',1001);	//try to get the small icon
			
			if(mh){
				
				WinDrawBitmap(MemHandleLock(mh),bounds->topLeft.x+1,bounds->topLeft.y+1);
				MemHandleUnlock(mh);
				DmReleaseResource(mh);
				
			}
			
			DmCloseDatabase(db);
		}
	}
	
	FntSetFont(font);
}


#define TIME_CONST_ONE_DAY	(24L*60L*60L)		//one day in seconds

Boolean handleSharewareShit(){				//return true if we need to show ui, false to just quit
	
	UInt32 rs;
	char str[64];
	char id[16];
	UInt16 bp=0;
	
	
	while(((rs=dcGetRegState())!=0xFFFFFFFF) && (bp==0)){
		
		StrIToA(id,dcGetUniqID());
		
		StrIToA(str,(rs+TIME_CONST_ONE_DAY-1)/TIME_CONST_ONE_DAY);
		
		bp=FrmCustomResponseAlert(1000,str,id,NULL,str,64,NULL);
		
		dcSetNewKey(StrAToI(str));
			
	}
	
	return rs!=0;
}

void UI(){
	
	FormPtr fp;
	ListPtr L;
	
	
	if(!handleSharewareShit())
		return;
	
	
	gNumListItems=listAllApps(&gListItems);		//yes, I know these are globals, but this is more readable and shows that they are affected
	
	fp=FrmInitForm(1000);
	L=FrmGetObjectPtr(fp,FrmGetObjectIndex(fp,1000));	//get list ptr
	LstSetDrawFunction(L,&listDrawFunc);
	LstSetListChoices(L,(char**)gListItems,gNumListItems);
	FrmSetEventHandler(fp,&mainFormEventHandler);
	FrmDoDialog(fp);
	FrmDeleteForm(fp);
	
	MemPtrFree(gListItems);						//free the items array
}


Int16 listItemSortCompareFunction(ListItem *i1,ListItem *i2,Int32 other){
	
	Int16 r=StrCompare(i1->name,i2->name);
	
	if(!r){		//same name? card comes second
		
		r=((i2->card)?-1:1);
		
	}
	
	return r;
}

UInt32 listAllApps(ListItem** items){			//return number of apps
	
	MemHandle mh=MemHandleNew(1);
	UInt16 count,i,j=0;
	ListItem* L;
	
	
	WinDrawChars(strEnumerating,StrLen(strEnumerating),0,0);
	//first we list the internal apps
	{
		MemHandle dbsH;
		SysDBListItemType* dbs;
		UInt16 dbAttr;
		
		
		SysCreateDataBaseList('appl',0,&count,&dbsH,true);
		dbs=MemHandleLock(dbsH);
		for(i=0;i<count;i++){
		
			DmDatabaseInfo(dbs[i].cardNo,dbs[i].dbID,NULL,&dbAttr,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);
			if(!(dbAttr&dmHdrAttrHidden)){
				MemHandleResize(mh,sizeof(ListItem)*(j+1));
				L=MemHandleLock(mh);
				
				StrCopy(L[j].name,dbs[i].name);
				L[j].crid=dbs[i].creator;
				L[j].card=false;
				
				MemHandleUnlock(mh);
				
				j++;
				
				{
					char x[12];
					StrIToA(x,j);
					WinDrawChars(x,StrLen(x),0,12);
				}
				
				
			}
			
		}
		MemHandleUnlock(dbsH);
		MemHandleFree(dbsH);
	}
	
	//now list VFS "/Palm/Launcher" and "/applications/  directories
	{
		UInt32 volIterator=0;
		UInt16 VRN=0;
		UInt16 idx=0;
		
		while(!VFSVolumeEnumerate(&VRN,&volIterator)){	//there are volumes to enumerate, we use the first
			
			FileRef cardFolder;
			
			while(StrLen(launcherFolderName[idx])!=0){
				
				if(!VFSFileOpen(VRN,launcherFolderName[idx],vfsModeRead,&cardFolder)){		//if launcher folder opened sucessfully
					UInt32 fileIterator=0;
					FileInfoType fi;
					char name[256];
					
					fi.nameP=name;
					fi.nameBufLen=256;
					
					while(!VFSDirEntryEnumerate(cardFolder,&fileIterator,&fi)){
					 	
					 	if(!(fi.attributes & (vfsFileAttrDirectory | vfsFileAttrLink))){	//if not folder and not link
							
							FileRef file;
							char longName[256];
							
							StrCopy(longName,launcherFolderName[idx]);
							StrCat(longName,name);
							StrCopy(name,longName);
							
							if(!VFSFileOpen(VRN,name,vfsModeRead,&file)){					//file opened
								
								UInt32 type,crid;
								UInt16 dbAttr;
								
								if(!VFSFileDBInfo(file,name,&dbAttr,NULL,NULL,NULL,NULL,NULL,NULL,NULL,&type,&crid,NULL)){
									
									if(type=='appl' && !(dbAttr&dmHdrAttrHidden)){
													//allright! we finally are sure we have a non-hidden palm APPLICATION,
													//   let's add it to the list
										MemHandleResize(mh,sizeof(ListItem)*(j+1));
										L=MemHandleLock(mh);
										
										StrCopy(L[j].name,name);
										L[j].crid=crid;
										L[j].card=true;
										
										MemHandleUnlock(mh);
										j++;
										
										{
											char x[12];
											StrIToA(x,j);
											WinDrawChars(x,StrLen(x),0,12);
										}
									}
								}	
								VFSFileClose(file);	
							}
						}
					}
					VFSFileClose(cardFolder);
				}
				idx++;
			}
		}
	}
	//add system default
	
	MemHandleResize(mh,sizeof(ListItem)*(j+1));
	L=MemHandleLock(mh);
									
	StrCopy(L[j].name,"  System Default");
	L[j].crid=SYSTEM_CRID;
	L[j].card=false;
	
	j++;
	
	SysQSort(L,j,sizeof(ListItem),(CmpFuncPtr)&listItemSortCompareFunction,0);
	
	
	//copy to output variable
	*items=MemPtrNew(sizeof(ListItem)*j);
	
	if(!(*items))
		SysFatalAlert("Out of memory enumerating apps");
	
	MemMove(*items,L,sizeof(ListItem)*j);
	
	MemHandleUnlock(mh);
	MemHandleFree(mh);
	
	return j;
}